from django.db import migrations


def forwards_encrypted_char(apps, schema_editor):
    EncryptedCharModel = apps.get_model("fields", "EncryptedCharModel")

    for row in EncryptedCharModel.objects.all():
        row.field = row.old_field
        row.save(update_fields=["field"])


def reverse_encrypted_char(apps, schema_editor):
    EncryptedCharModel = apps.get_model("fields", "EncryptedCharModel")

    for row in EncryptedCharModel.objects.all():
        row.old_field = row.field
        row.save(update_fields=["old_field"])


def forwards_encrypted_date_time(apps, schema_editor):
    EncryptedDateTimeModel = apps.get_model("fields", "EncryptedDateTimeModel")

    for row in EncryptedDateTimeModel.objects.all():
        row.auto_now = row.old_auto_now
        row.date = row.old_date
        row.datetime = row.old_datetime
        row.time = row.old_time
        row.save(update_fields=["auto_now", "date", "datetime", "time"])


def reverse_encrypted_date_time(apps, schema_editor):
    EncryptedDateTimeModel = apps.get_model("fields", "EncryptedDateTimeModel")

    for row in EncryptedDateTimeModel.objects.all():
        row.old_auto_now = row.auto_now
        row.old_date = row.date
        row.old_datetime = row.datetime
        row.old_time = row.time
        row.save(update_fields=["old_auto_now", "old_date", "old_datetime", "old_time"])


def forwards_encrypted_integer(apps, schema_editor):
    EncryptedIntegerModel = apps.get_model("fields", "EncryptedIntegerModel")

    for row in EncryptedIntegerModel.objects.all():
        row.field = row.old_field
        row.save(update_fields=["field"])


def reverse_encrypted_integer(apps, schema_editor):
    EncryptedIntegerModel = apps.get_model("fields", "EncryptedIntegerModel")

    for row in EncryptedIntegerModel.objects.all():
        row.old_field = row.field
        row.save(update_fields=["old_field"])


def forwards_encrypted_nullable_integer(apps, schema_editor):
    EncryptedNullableIntegerModel = apps.get_model(
        "fields", "EncryptedNullableIntegerModel"
    )

    for row in EncryptedNullableIntegerModel.objects.all():
        row.field = row.old_field
        row.save(update_fields=["field"])


def reverse_encrypted_nullable_integer(apps, schema_editor):
    EncryptedNullableIntegerModel = apps.get_model(
        "fields", "EncryptedNullableIntegerModel"
    )

    for row in EncryptedNullableIntegerModel.objects.all():
        row.old_field = row.field
        row.save(update_fields=["old_field"])


def forwards_encrypted_ttl_integer(apps, schema_editor):
    EncryptedTTLIntegerModel = apps.get_model("fields", "EncryptedTTLIntegerModel")

    for row in EncryptedTTLIntegerModel.objects.all():
        row.field = row.old_field
        row.save(update_fields=["field"])


def reverse_encrypted_ttl_integer(apps, schema_editor):
    EncryptedTTLIntegerModel = apps.get_model("fields", "EncryptedTTLIntegerModel")

    for row in EncryptedTTLIntegerModel.objects.all():
        row.old_field = row.field
        row.save(update_fields=["old_field"])


def forwards_other_encrypted_types(apps, schema_editor):
    OtherEncryptedTypesModel = apps.get_model("fields", "OtherEncryptedTypesModel")

    for row in OtherEncryptedTypesModel.objects.all():
        row.decimal = row.old_decimal
        row.ip = row.old_ip
        row.uuid = row.old_uuid
        row.save(update_fields=["decimal", "ip", "uuid"])


def reverse_other_encrypted_types(apps, schema_editor):
    OtherEncryptedTypesModel = apps.get_model("fields", "OtherEncryptedTypesModel")

    for row in OtherEncryptedTypesModel.objects.all():
        row.old_decimal = row.decimal
        row.old_ip = row.ip
        row.old_uuid = row.uuid
        row.save(update_fields=["old_decimal", "old_ip", "old_uuid"])


class Migration(migrations.Migration):
    dependencies = [
        ("fields", "0003_add_encrypted_fields"),
    ]

    operations = [
        migrations.RunPython(forwards_encrypted_char, reverse_encrypted_char),
        migrations.RunPython(forwards_encrypted_date_time, reverse_encrypted_date_time),
        migrations.RunPython(forwards_encrypted_integer, reverse_encrypted_integer),
        migrations.RunPython(
            forwards_encrypted_nullable_integer, reverse_encrypted_nullable_integer
        ),
        migrations.RunPython(
            forwards_encrypted_ttl_integer, reverse_encrypted_ttl_integer
        ),
        migrations.RunPython(
            forwards_other_encrypted_types, reverse_other_encrypted_types
        ),
    ]



import decimal
import json
import uuid
from io import StringIO

from django import forms
from django.conf import settings
from django.core import exceptions, serializers, validators
from django.core.management import call_command
from django.db import IntegrityError, connection, models
from django.test import TestCase, TransactionTestCase, override_settings
from django.test.utils import freeze_time
from django.utils import timezone

from django_cryptography.fields import Expired, encrypt

from .models import (
    EncryptedCharModel,
    EncryptedDateTimeModel,
    EncryptedFieldSubclass,
    EncryptedIntegerModel,
    EncryptedNullableIntegerModel,
    EncryptedTTLIntegerModel,
    OtherEncryptedTypesModel,
)


class TestSaveLoad(TestCase):
    def test_integer(self):
        instance = EncryptedIntegerModel(field=42)
        instance.save()
        loaded = EncryptedIntegerModel.objects.get()
        self.assertEqual(instance.field, loaded.field)

    def test_char(self):
        instance = EncryptedCharModel(field="Hello, world!")
        instance.save()
        loaded = EncryptedCharModel.objects.get()
        self.assertEqual(instance.field, loaded.field)

    def test_dates(self):
        instance = EncryptedDateTimeModel(
            datetime=timezone.now(),
            date=timezone.now().date(),
            time=timezone.now().time(),
        )
        instance.save()
        loaded = EncryptedDateTimeModel.objects.get()
        self.assertEqual(instance.datetime, loaded.datetime)
        self.assertEqual(instance.date, loaded.date)
        self.assertEqual(instance.time, loaded.time)
        self.assertTrue(instance.auto_now)
        self.assertEqual(instance.auto_now, loaded.auto_now)

    def test_default_null(self):
        instance = EncryptedNullableIntegerModel()
        instance.save()
        loaded = EncryptedNullableIntegerModel.objects.get(pk=instance.pk)
        self.assertEqual(loaded.field, None)
        self.assertEqual(instance.field, loaded.field)

    def test_null_handling(self):
        instance = EncryptedNullableIntegerModel(field=None)
        instance.save()
        loaded = EncryptedNullableIntegerModel.objects.get()
        self.assertEqual(instance.field, loaded.field)

        instance = EncryptedIntegerModel(field=None)
        with self.assertRaises(IntegrityError):
            instance.save()

    def test_ttl(self):
        with freeze_time(499162800):
            instance = EncryptedTTLIntegerModel(field=42)
            instance.save()

        with freeze_time(123456789):
            loaded = EncryptedTTLIntegerModel.objects.get()
            self.assertIs(loaded.field, Expired)

    def test_other_types(self):
        instance = OtherEncryptedTypesModel(
            ip="192.168.0.1",
            uuid=uuid.uuid4(),
            decimal=decimal.Decimal(1.25),
        )
        instance.save()
        loaded = OtherEncryptedTypesModel.objects.get()
        self.assertEqual(instance.ip, loaded.ip)
        self.assertEqual(instance.uuid, loaded.uuid)
        self.assertEqual(instance.decimal, loaded.decimal)

    def test_updates(self):
        with self.assertNumQueries(2):
            instance = EncryptedCharModel.objects.create(field="Hello, world!")
            instance.field = "Goodbye, world!"
            instance.save()
        loaded = EncryptedCharModel.objects.get()
        self.assertEqual(instance.field, loaded.field)


class TestQuerying(TestCase):
    def setUp(self):
        self.objs = [
            EncryptedNullableIntegerModel.objects.create(field=1),
            EncryptedNullableIntegerModel.objects.create(field=2),
            EncryptedNullableIntegerModel.objects.create(field=3),
            EncryptedNullableIntegerModel.objects.create(field=None),
        ]

    def test_isnull(self):
        self.assertSequenceEqual(
            self.objs[-1:],
            EncryptedNullableIntegerModel.objects.filter(field__isnull=True),
        )

    def test_unsupported(self):
        with self.assertRaises(exceptions.FieldError):
            EncryptedNullableIntegerModel.objects.filter(field__exact=2)


class TestChecks(TestCase):
    def test_settings_has_key(self):
        key = settings.CRYPTOGRAPHY_KEY
        self.assertIsNotNone(key)
        self.assertIsInstance(key, bytes)

    def test_field_description(self):
        field = encrypt(models.IntegerField())
        self.assertEqual("Encrypted Integer", field.description)

    def test_field_checks(self):
        class BadField(models.Model):
            field = encrypt(models.CharField())

            class Meta:
                app_label = "myapp"

        model = BadField()
        errors = model.check()
        self.assertEqual(len(errors), 1)
        # The inner CharField is missing a max_length.
        self.assertEqual("fields.E120", errors[0].id)
        self.assertIn("max_length", errors[0].msg)

    def test_invalid_base_fields(self):
        class Related(models.Model):
            field = encrypt(
                models.ForeignKey("fields.EncryptedIntegerModel", models.CASCADE)
            )

            class Meta:
                app_label = "myapp"

        obj = Related()
        errors = obj.check()
        self.assertEqual(1, len(errors))
        self.assertEqual("encrypted.E002", errors[0].id)


class TestMigrations(TransactionTestCase):
    available_apps = ["tests.fields"]

    def test_clone(self):
        field = encrypt(models.IntegerField())
        new_field = field.clone()
        self.assertIsNot(field, new_field)
        self.assertEqual(field.verbose_name, new_field.verbose_name)
        self.assertNotEqual(field.creation_counter, new_field.creation_counter)

    def test_subclass_clone(self):
        field = EncryptedFieldSubclass()
        new_field = field.clone()
        self.assertIsNot(field, new_field)
        self.assertEqual(field.verbose_name, new_field.verbose_name)
        self.assertNotEqual(field.creation_counter, new_field.creation_counter)

    def test_deconstruct(self):
        field = encrypt(models.IntegerField())
        name, path, args, kwargs = field.deconstruct()
        new = encrypt(*args, **kwargs)
        self.assertEqual(type(new), type(field))

    def test_deconstruct_with_ttl(self):
        field = encrypt(models.IntegerField(), ttl=60)
        name, path, args, kwargs = field.deconstruct()
        new = encrypt(*args, **kwargs)
        self.assertEqual(new.ttl, field.ttl)

    def test_deconstruct_args(self):
        field = encrypt(models.CharField(max_length=20))
        name, path, args, kwargs = field.deconstruct()
        new = encrypt(*args, **kwargs)
        self.assertEqual(new.max_length, field.max_length)

    def test_subclass_deconstruct(self):
        field = encrypt(models.IntegerField())
        name, path, args, kwargs = field.deconstruct()
        self.assertEqual("django_cryptography.fields.encrypt", path)

        field = EncryptedFieldSubclass()
        name, path, args, kwargs = field.deconstruct()
        self.assertEqual("tests.fields.models.EncryptedFieldSubclass", path)

    @override_settings(
        MIGRATION_MODULES={"fields": "tests.fields.test_migrations_encrypted_default"}
    )
    def test_adding_field_with_default(self):
        table_name = "fields_integerencrypteddefaultmodel"
        with connection.cursor() as cursor:
            self.assertNotIn(table_name, connection.introspection.table_names(cursor))
        call_command("migrate", "fields", verbosity=0)
        with connection.cursor() as cursor:
            self.assertIn(table_name, connection.introspection.table_names(cursor))
        call_command("migrate", "fields", "zero", verbosity=0)
        with connection.cursor() as cursor:
            self.assertNotIn(table_name, connection.introspection.table_names(cursor))

    @override_settings(
        MIGRATION_MODULES={"fields": "tests.fields.test_migrations_normal_to_encrypted"}
    )
    def test_makemigrations_no_changes(self):
        out = StringIO()
        call_command("makemigrations", "--dry-run", "fields", stdout=out)
        self.assertIn("No changes detected in app 'fields'", out.getvalue())


class TestSerialization(TestCase):
    test_data_integer = (
        '[{"fields": {"field": 42}, '
        '"model": "fields.encryptedintegermodel", "pk": null}]'
    )
    test_data_char = (
        '[{"fields": {"field": "Hello, world!"}, '
        '"model": "fields.encryptedcharmodel", "pk": null}]'
    )

    def test_integer_dumping(self):
        instance = EncryptedIntegerModel(field=42)
        data = serializers.serialize("json", [instance])
        self.assertEqual(json.loads(self.test_data_integer), json.loads(data))

    def test_integer_loading(self):
        instance = list(serializers.deserialize("json", self.test_data_integer))[
            0
        ].object
        self.assertEqual(42, instance.field)

    def test_char_dumping(self):
        instance = EncryptedCharModel(field="Hello, world!")
        data = serializers.serialize("json", [instance])
        self.assertEqual(json.loads(self.test_data_char), json.loads(data))

    def test_char_loading(self):
        instance = list(serializers.deserialize("json", self.test_data_char))[0].object
        self.assertEqual("Hello, world!", instance.field)


class TestValidation(TestCase):
    def test_unbounded(self):
        field = encrypt(models.IntegerField())
        with self.assertRaises(exceptions.ValidationError) as cm:
            field.clean(None, None)
        self.assertEqual("null", cm.exception.code)
        self.assertEqual("This field cannot be null.", cm.exception.messages[0])

    def test_blank_true(self):
        field = encrypt(models.IntegerField(blank=True, null=True))
        # This should not raise a validation error
        field.clean(None, None)

    def test_with_validators(self):
        field = encrypt(
            models.IntegerField(validators=[validators.MinValueValidator(1)])
        )
        field.clean(1, None)
        with self.assertRaises(exceptions.ValidationError) as cm:
            field.clean(0, None)
        self.assertEqual(
            "Ensure this value is greater than or equal to 1.", cm.exception.messages[0]
        )


class TestFormField(TestCase):
    class EncryptedCharModelForm(forms.ModelForm):
        class Meta:
            fields = "__all__"
            model = EncryptedCharModel

    def test_model_field_formfield(self):
        model_field = encrypt(models.CharField(max_length=27))
        form_field = model_field.formfield()
        self.assertIsInstance(form_field, forms.CharField)
        self.assertEqual(form_field.max_length, 27)

    def test_model_form(self):
        data = {"field": "Hello, world!"}
        form = self.EncryptedCharModelForm(data)
        self.assertTrue(form.is_valid(), form.errors)
        self.assertEqual({"field": "Hello, world!"}, form.cleaned_data)

        instance = form.save()
        loaded = EncryptedCharModel.objects.get()
        self.assertEqual(instance.field, loaded.field)

    def test_model_form_update(self):
        data = {"field": "Goodbye, world!"}
        instance = EncryptedCharModel.objects.create(field="Hello, world!")
        form = self.EncryptedCharModelForm(data, instance=instance)
        self.assertTrue(form.is_valid(), form.errors)
        form.save()

        loaded = EncryptedCharModel.objects.get()
        self.assertEqual(data["field"], loaded.field)



from typing import Any, Dict

from appconf import AppConf
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.kdf import pbkdf2
from django.conf import settings
from django.utils.encoding import force_bytes


class CryptographyConf(AppConf):
    BACKEND = default_backend()
    DIGEST = hashes.SHA256()
    KEY = None
    SALT = "django-cryptography"

    class Meta:
        prefix = "cryptography"
        proxy = True

    def configure_salt(self, value: Any) -> bytes:
        return force_bytes(value)

    def configure(self) -> Dict[str, Any]:
        backend = self.configured_data["BACKEND"]
        digest = self.configured_data["DIGEST"]
        salt = self.configured_data["SALT"]
        # Key Derivation Function
        kdf = pbkdf2.PBKDF2HMAC(
            algorithm=digest,
            length=digest.digest_size,
            salt=salt,
            iterations=30000,
            backend=backend,
        )
        self.configured_data["KEY"] = kdf.derive(
            force_bytes(self.configured_data["KEY"] or settings.SECRET_KEY)
        )
        return self.configured_data



import pickle
from base64 import b64decode, b64encode
from typing import (
    Any,
    Dict,
    List,
    Optional,
    Sequence,
    Tuple,
    Type,
    TypeVar,
    Union,
    cast,
    overload,
)

from django.core import checks
from django.core.checks import CheckMessage
from django.db import models
from django.db.backends.base.base import BaseDatabaseWrapper
from django.db.models.lookups import Lookup, Transform
from django.utils.encoding import force_bytes
from django.utils.translation import gettext_lazy as _

from django_cryptography.core.signing import SignatureExpired
from django_cryptography.typing import DatabaseWrapper
from django_cryptography.utils.crypto import FernetBytes

F = TypeVar("F", bound=models.Field)
FIELD_CACHE: Dict[type, type] = {}

Expired = object()
"""Represents an expired encryption value."""


class PickledField(models.BinaryField):
    """
    A field for storing pickled objects
    """

    description = _("Pickled data")
    empty_values = [None, b""]
    supported_lookups = ("exact", "in", "isnull")

    def _dump(self, value: Any) -> bytes:
        return pickle.dumps(value)

    def _load(self, value: bytes) -> Any:
        return pickle.loads(value)

    def get_lookup(self, lookup_name: str) -> Optional[Type[Lookup]]:
        if lookup_name not in self.supported_lookups:
            return None
        return super().get_lookup(lookup_name)

    def get_transform(self, lookup_name: str) -> Optional[Type[Transform]]:
        if lookup_name not in self.supported_lookups:
            return None
        return super().get_transform(lookup_name)

    def get_db_prep_value(
        self, value: Any, connection: BaseDatabaseWrapper, prepared: bool = False
    ) -> Optional[bytes]:
        if value is not None:
            value = self._dump(value)
        return super().get_db_prep_value(value, connection, prepared)

    def from_db_value(self, value: Any, *args: Any, **kwargs: Any) -> Any:
        if value is not None:
            return self._load(force_bytes(value))
        return value

    def value_to_string(self, obj: models.Model) -> str:
        """Pickled data is serialized as base64"""
        return b64encode(self._dump(self.value_from_object(obj))).decode("ascii")

    def to_python(self, value: Optional[Any]) -> Optional[Any]:
        # If it's a string, it should be base64-encoded data
        if isinstance(value, str):
            return self._load(b64decode(force_bytes(value)))
        return value


class EncryptedMixin(models.Field):
    """
    A field mixin storing encrypted data

    :param bytes key: This is an optional argument.

        Allows for specifying an instance specific encryption key.
    :param int ttl: This is an optional argument.

        The amount of time in seconds that a value can be stored for. If the
        time to live of the data has passed, it will become unreadable.
        The expired value will return an :class:`Expired` object.
    """

    supported_lookups = ("isnull",)

    def __init__(self, *args, **kwargs) -> None:
        self.base_class: Type[models.Field]
        self.wasinstance: bool

        self.key: Union[bytes, str] = kwargs.pop("key", None)
        self.ttl: int = kwargs.pop("ttl", None)

        self._fernet = FernetBytes(self.key)
        super().__init__(*args, **kwargs)

    def _description(self) -> str:
        return _("Encrypted %s") % super().description

    description = property(_description)  # type: ignore[assignment]

    def _dump(self, value: Any) -> bytes:
        return self._fernet.encrypt(pickle.dumps(value))

    def _load(self, value: bytes) -> Any:
        try:
            return pickle.loads(self._fernet.decrypt(value, self.ttl))
        except SignatureExpired:
            return Expired

    def check(self, **kwargs: Any) -> List[CheckMessage]:
        errors = super().check(**kwargs)
        if getattr(self, "remote_field", None):
            errors.append(
                checks.Error(
                    "Base field for encrypted cannot be a related field.",
                    hint=None,
                    obj=self,
                    id="encrypted.E002",
                )
            )
        return errors

    def clone(self):
        name, path, args, kwargs = super().deconstruct()
        # Determine if the class that subclassed us has been subclassed.
        if self.__class__.__mro__.index(EncryptedMixin) <= 1:
            return encrypt(self.base_class(*args, **kwargs), self.key, self.ttl)
        return self.__class__(*args, **kwargs)

    def deconstruct(self) -> Tuple[str, str, Sequence[Any], Dict[str, Any]]:
        name, path, args, kwargs = super().deconstruct()
        if self.wasinstance is False:
            path = f"{self.base_class.__module__}.{self.base_class.__name__}"
        # Determine if the class that subclassed us has been subclassed.
        elif self.__class__.__mro__.index(EncryptedMixin) <= 1:
            path = f"{encrypt.__module__}.{encrypt.__name__}"
            args = [self.base_class(*args, **kwargs)]
            kwargs = {}
            if self.ttl is not None:
                kwargs["ttl"] = self.ttl
        return name, path, args, kwargs

    def get_lookup(self, lookup_name: str) -> Optional[Any]:
        if lookup_name not in self.supported_lookups:
            return None
        return super().get_lookup(lookup_name)

    def get_transform(self, lookup_name: str) -> Optional[Any]:
        if lookup_name not in self.supported_lookups:
            return None
        return super().get_transform(lookup_name)

    def get_internal_type(self) -> str:
        return "BinaryField"

    def get_db_prep_value(
        self, value: Any, connection: BaseDatabaseWrapper, prepared: bool = False
    ) -> Any:
        value = models.Field.get_db_prep_value(self, value, connection, prepared)
        if value is not None:
            return cast(DatabaseWrapper, connection).Database.Binary(self._dump(value))
        return value

    get_db_prep_save = models.Field.get_db_prep_save

    def from_db_value(self, value, *args, **kwargs) -> Any:
        if value is not None:
            return self._load(force_bytes(value))
        return value


def get_encrypted_field(base_class: Type[F], wasinstance: bool) -> Type[F]:
    """
    A get or create method for encrypted fields, we cache the field in
    the module to avoid recreation. This also allows us to always return
    the same class reference for a field.
    """
    assert issubclass(base_class, models.Field)
    return FIELD_CACHE.setdefault(
        base_class,
        type(
            ("Encrypted" if wasinstance else "") + base_class.__name__,
            (EncryptedMixin, base_class),
            {"base_class": base_class, "wasinstance": wasinstance},
        ),
    )


@overload
def encrypt(base_field: F, key=None, ttl=None) -> F:
    ...


@overload
def encrypt(base_field: Type[F]) -> Type[F]:
    ...


def encrypt(
    base_field, key: Optional[Union[bytes, str]] = None, ttl: Optional[int] = None
):
    """
    A decorator for creating encrypted model fields.

    :param base_field: Base Field to encrypt
    :param bytes key: This is an optional argument.

        Allows for specifying an instance specific encryption key.
    :param int ttl: This is an optional argument.

        The amount of time in seconds that a value can be stored for. If the
        time to live of the data has passed, it will become unreadable.
        The expired value will return an :class:`Expired` object.
    """
    if isinstance(base_field, type):
        assert issubclass(base_field, models.Field)
        assert key is None
        assert ttl is None
        return get_encrypted_field(base_field, False)

    name, path, args, kwargs = base_field.deconstruct()
    kwargs.update({"key": key, "ttl": ttl})
    return get_encrypted_field(type(base_field), True)(*args, **kwargs)
